import abc
from datetime import datetime
from time import perf_counter
from typing import Callable, TypedDict, NotRequired, Type, Any, Awaitable, Self

import wrapt


class FunctionCall(TypedDict):
    timestamp: float
    method: str
    args: list[str]
    kwargs: dict[str, str]
    execution_time: NotRequired[float]
    result: NotRequired[str]


Exporter = Callable[[FunctionCall], Awaitable[None]]


class InMemoryExporter:
    def __init__(self):
        self.events = []

    async def export(self, events: FunctionCall) -> None:
        self.events.append(events)


class HarambeInstrumentation(abc.ABC):
    __WRAPPED_FUNCTIONS = dict()

    PLAYWRIGHT_METHODS_TO_INSTRUMENT = [
        "click",
        "goto",
        "query_selector",
        "query_selector_all",
        "wait_for_selector",
        "wait_for_load_state",
        "wait_for_timeout",
        "screenshot",
        "evaluate",
    ]

    SOUP_METHODS_TO_INSTRUMENT = [
        "goto",
        "query_selector",
        "query_selector_all",
    ]

    SDK_METHODS_TO_INSTRUMENT = [
        "enqueue",
        "save_data",
        "save_cookies",
        "save_local_storage",
        "capture_pdf",
        "capture_url",
        "capture_html",
        "capture_download",
    ]

    def __init__(self):
        self._exporters = []

    def add_exporter(self, exporter: Exporter) -> Self:
        self._exporters.append(exporter)
        return self

    async def export(self, event: FunctionCall) -> None:
        for exporter in self._exporters:
            await exporter(event)

    def instrument(self) -> Self:
        from playwright.async_api import Page
        from harambe.contrib.soup.impl import SoupPage
        from harambe import SDK

        for method_name in self.PLAYWRIGHT_METHODS_TO_INSTRUMENT:
            self._wrap_function(Page, method_name)

        for method_name in self.SOUP_METHODS_TO_INSTRUMENT:
            self._wrap_function(SoupPage, method_name)

        for method_name in self.SDK_METHODS_TO_INSTRUMENT:
            self._wrap_function(SDK, method_name)

        return self

    @staticmethod
    def un_instrument() -> None:
        for (
            target,
            method_name,
        ), func in HarambeInstrumentation.__WRAPPED_FUNCTIONS.items():
            setattr(target, method_name, func)

        HarambeInstrumentation.__WRAPPED_FUNCTIONS = dict()

    def _wrap_function(self, target: Type[Any], method_name: str):
        target_name = target.__name__

        if HarambeInstrumentation.__WRAPPED_FUNCTIONS.get((target, method_name)):
            print("Already instrumented: ", target_name + "." + method_name)
            return

        HarambeInstrumentation.__WRAPPED_FUNCTIONS[(target, method_name)] = getattr(
            target, method_name
        )

        async def _wrapper(func, _instance, args, kwargs):
            event: FunctionCall = {
                "timestamp": datetime.now().timestamp(),
                "method": f"{target_name}.{method_name}",
                "args": [str(a) for a in args],
                "kwargs": {k: repr(v) for k, v in kwargs.items()},
            }

            exc, result = None, None
            start = perf_counter()

            try:
                result = await func(*args, **kwargs)
                event["result"] = repr(result)

            except Exception as e:
                event["result"] = repr(e)
                exc = e

            event["execution_time"] = perf_counter() - start
            await self.export(event)

            if exc:
                raise exc
            return result

        wrapt.wrap_function_wrapper(target, method_name, _wrapper)



from harambe.cache import single_value_cache


async def test_single_value_cache(mocker):
    mock_fn = mocker.AsyncMock(return_value="result")

    class Dummy:
        def __init__(self):
            self._cache = None

        @single_value_cache("_cache")
        async def f(self, key):
            return await mock_fn(key)

    c = Dummy()

    # first call: cache miss
    r1 = await c.f("a")
    assert r1 == "result"
    assert mock_fn.await_count == 1

    # second call same key: cache hit
    r2 = await c.f("a")
    assert r2 == "result"
    assert mock_fn.await_count == 1

    # new key: cache miss
    r3 = await c.f("b")
    assert r3 == "result"
    assert mock_fn.await_count == 2

    # again "b": hit
    r4 = await c.f("b")
    assert r4 == "result"
    assert mock_fn.await_count == 2

    # new key: cache miss
    r5 = await c.f("a")
    assert r5 == "result"

    assert mock_fn.await_count == 3

    # new key: cache miss
    r6 = await c.f("c")
    assert r6 == "result"

    assert mock_fn.await_count == 4



import pytest

from .matchers import assert_partial_object_in


def test_assert_partial_object_in_with_matching_object():
    collection = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    partial_obj = {"id": 1}
    assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_no_matching_object():
    collection = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    partial_obj = {"id": 3}
    with pytest.raises(AssertionError):
        assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_multiple_keys_match():
    collection = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    partial_obj = {"id": 1, "name": "Alice"}
    assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_partial_keys_mismatch():
    collection = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    partial_obj = {"id": 1, "name": "Bob"}
    with pytest.raises(AssertionError):
        assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_empty_collection():
    collection = []
    partial_obj = {"id": 1}
    with pytest.raises(AssertionError):
        assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_empty_partial_obj():
    collection = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    partial_obj = {}
    assert_partial_object_in(collection, partial_obj)


def test_assert_partial_object_in_with_non_dict_collection():
    collection = ({"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"})
    partial_obj = {"id": 1}
    assert_partial_object_in(collection, partial_obj)



from functools import wraps
from typing import Any


def single_value_cache(cache_attr: str):
    """
    Decorator factory to cache a single (key, value) tuple on `self`.
    :param cache_attr: name of the attribute on self that holds the cache tuple.
    """

    def decorator(fn):
        @wraps(fn)
        async def wrapper(self, key: Any, *args, **kwargs):
            cache = getattr(self, cache_attr, None)
            if cache is not None and cache[0] == key:
                return cache[1]

            result = await fn(self, key, *args, **kwargs)
            setattr(self, cache_attr, (key, result))

            return result

        return wrapper

    return decorator



import datetime
from unittest.mock import AsyncMock

import pytest
from playwright.async_api import Page

from harambe.contrib.soup.impl import SoupPage
from harambe.core import SDK
from harambe.instrumentation import HarambeInstrumentation, InMemoryExporter


@pytest.fixture(autouse=True)
def patch_perf_counter_and_datetime(mocker):
    mocker.patch("harambe.instrumentation.perf_counter", side_effect=[0.0, 1.0] * 100)
    dt = mocker.patch("harambe.instrumentation.datetime")
    dt.now.return_value = datetime.datetime(1997, 3, 8, tzinfo=datetime.UTC)


@pytest.fixture
def exporter():
    return InMemoryExporter()


@pytest.fixture(
    params=[
        Page(
            impl_obj=AsyncMock(
                url="https://example.com", query_selector=AsyncMock(return_value=None)
            )
        ),
        SoupPage(session=AsyncMock(), url="https://example.com"),
    ]
)
def sdk(request, exporter):
    instrumentation = (
        HarambeInstrumentation().add_exporter(exporter.export).instrument()
    )
    yield SDK(page=request.param)
    instrumentation.un_instrument()


async def test_successful_method_call(sdk, exporter):
    await sdk.save_data({"test": "data"})

    assert len(exporter.events) == 2
    event = exporter.events[0]

    page_cls = sdk.page.__class__.__name__
    assert event["method"] == f"{page_cls}.query_selector"
    assert event["args"] == ["base"]
    assert event["kwargs"] == {}
    assert event["result"] == "None"
    assert event["timestamp"] == 857779200.0

    event = exporter.events[1]
    assert event["method"] == "SDK.save_data"
    assert event["args"] == ["{'test': 'data'}"]
    assert event["kwargs"] == {}
    assert event["result"] == "None"
    assert event["timestamp"] == 857779200.0
    assert event["execution_time"] == 1.0


async def test_method_call_with_exception(sdk, exporter):
    with pytest.raises(ValueError) as e:
        await sdk.capture_html("div.content")

    # Verify the event was recorded
    assert len(exporter.events) == 2
    page_cls = sdk.page.__class__.__name__
    assert exporter.events[0]["method"] == f"{page_cls}.query_selector"
    assert exporter.events[0]["result"] == "None"

    event = exporter.events[1]
    assert event["method"] == "SDK.capture_html"
    assert event["args"] == ["div.content"]
    assert event["kwargs"] == {}
    assert (
        event["result"] == "ValueError('Element not found for selector: div.content')"
    )
    assert event["timestamp"] == 857779200.0
    assert event["execution_time"] == 1.0


async def test_method_call_with_return_value(sdk, exporter):
    sdk.page.pdf = AsyncMock(return_value=b"pdf")

    result = await sdk.capture_pdf()

    expected_result = {
        "url": "https://example.com/reworkd_page_pdf.pdf",
        "filename": "reworkd_page_pdf.pdf",
    }
    assert result == expected_result

    event = exporter.events[-1]
    assert event["method"] == "SDK.capture_pdf"
    assert event["args"] == []
    assert event["kwargs"] == {}
    assert event["result"] == str(expected_result)
    assert event["timestamp"] == 857779200.0


async def test_method_with_args_and_kwargs(sdk, exporter):
    await sdk.enqueue(
        "https://example.com",
        "https://example.com/next",
        options={"waitUntil": "load"},
        context={"foo": "bar"},
    )

    assert len(exporter.events) == 2
    page_cls = sdk.page.__class__.__name__
    assert exporter.events[0]["method"] == f"{page_cls}.query_selector"

    event = exporter.events[1]
    assert event["method"] == "SDK.enqueue"
    assert event["args"] == ["https://example.com", "https://example.com/next"]
    assert event["kwargs"] == {
        "context": "{'foo': 'bar'}",
        "options": "{'waitUntil': 'load'}",
    }
    assert event["result"] == "None"
    assert event["timestamp"] == 857779200.0
    assert event["execution_time"] == 1.0
